﻿/*
 * This software released under the MIT License
 * https://opensource.org/licenses/MIT
 * */
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace clique_count
{
    public static class Global
    {
        private static object _lock = new object();

        /// <summary>
        /// Global counter to track the number of active threads
        /// </summary>
        public static int TotalWorkers { get; private set; }

        /// <summary>
        /// Thread safe worker addition.
        /// </summary>
        /// <param name="n"></param>
        public static void AddWorkers(int n)
        {
            lock (_lock)
            {
                TotalWorkers += n;
            }
        }

        /// <summary>
        /// Thread safe worker removal. A worker should call this function when it's done.
        /// </summary>
        /// <param name="n"></param>
        public static void RemoveWorkers(int n)
        {
            lock (_lock)
            {
                TotalWorkers -= n;
            }
        }

        /// <summary>
        /// Worker input will be split into two groups this many times, recursively.
        /// </summary>
        public static int MaxDepth { get; set; }

        public static bool Reflect { get; set; }
    }

    public class Program
    {
        static void Main(string[] args)
        {
            //UInt64 ttt = 1;
            //var x = Reverse(ttt, 4);
            //return;
            int startingNodes = 5;
            int nodes = 20;

            Global.Reflect = false;

            if (!Object.ReferenceEquals(null, args))
            {
                string exitMessage = "call with '[/r] x y' where x is starting nodes (>1) and y is max nodes (>starting nodes)";
                bool exitError = false;

                if (args.Length > 0 && !(args.Length == 2 || args.Length == 3))
                {
                    exitError = true;
                }
                
                if (args.Length == 2)
                {
                    if (!int.TryParse(args[0], out startingNodes))
                    {
                        exitError = true;
                    }

                    if (!int.TryParse(args[1], out nodes))
                    {
                        exitError = true;
                    }

                    if (startingNodes < 2 || nodes < 2 || startingNodes >= nodes)
                    {
                        exitError = true;
                    }
                }
                else if (args.Length == 3)
                {
                    if (!(args[0] == "/r" || args[0] == "/R"))
                    {
                        exitError = true;
                    }
                    else
                    {
                        Global.Reflect = true;
                    }

                    if (!int.TryParse(args[1], out startingNodes))
                    {
                        exitError = true;
                    }

                    if (!int.TryParse(args[2], out nodes))
                    {
                        exitError = true;
                    }

                    if (startingNodes < 2 || nodes < 2 || startingNodes >= nodes)
                    {
                        exitError = true;
                    }
                }
                
                if (exitError)
                {
                    System.Console.WriteLine(exitMessage);
                    System.Environment.Exit(0);
                }
            }

            if (Global.Reflect)
            {
                System.Console.WriteLine("Counting with reflection.");
            }
            
            var stopwatch = new Stopwatch();
            stopwatch.Start();

            for (int i=startingNodes; i<nodes; i++)
            //int i = 16;
            {
                System.Console.WriteLine(String.Format("FamilyCount({0},1..{1}): ", i,i/2+1));

                var counts = new List<UInt64>();
                var reverseStack = new Stack<UInt64>();

                for (int j = 1; j < (i / 2) + 1; j++)
                //int j = 8;
                {
                    if (j > 16)
                    {
                        Global.MaxDepth = 4;
                    }
                    else if (j > 8 && j <= 16)
                    {
                        Global.MaxDepth = 3;
                    }
                    else if (j > 4 && j <= 8)
                    {
                        Global.MaxDepth = 2;
                    }
                    else if (j <= 4)
                    {
                        Global.MaxDepth = 1;
                    }

                    int totalNodes = 1 << (Global.MaxDepth + 2);
                    var total = new UInt64[totalNodes];

                    for (int s = 0; s < totalNodes; s++)
                    {
                        total[s] = 0;
                    }

                    SplitWork(total, i, j);

                    while (Global.TotalWorkers > 0)
                    {
                        System.Threading.Thread.Sleep(50);
                    }

                    UInt64 sum = 0;
                    for (int s = 0; s < totalNodes; s++)
                    {
                        sum += total[s];
                    }

                    System.Console.Write(sum);

                    if (j < (i / 2))
                    {
                        System.Console.Write(",");

                    }

                    if (j < (i + 1) / 2)
                    {
                        reverseStack.Push(sum);
                    }
                }

                if (i > 2)
                {
                    System.Console.Write("," + String.Join(",", reverseStack.ToList()));
                    reverseStack.Clear();
                }

                System.Console.WriteLine();
            }

            stopwatch.Stop();

            System.Console.WriteLine("Elapsed: " + stopwatch.Elapsed.TotalSeconds.ToString());
        }

        // Top level method. Call this.
        static void SplitWork(UInt64[] total, int nodes, int cliques)
        {
            if (cliques < 2)
            {
                Global.AddWorkers(1);

                Task.Factory.StartNew(() => Worker(total, nodes, cliques, 0, 0, nodes));

                return;
            }

            SplitWorkRecursive(total, nodes, cliques, 0, 0, 1, 0, nodes);
        }

        /// <summary>
        /// Recursively splits the counting out into different workers. 
        /// </summary>
        /// <param name="total">Global container to store results. Each thread gets its own slot.</param>
        /// <param name="nodes">number of nodes.</param>
        /// <param name="cliques">number of cliques.</param>
        /// <param name="add">the "top half" of, e.g., C(20,7) is the same as C(19,7) with each value added to (1<<19). </param>
        /// <param name="startIndex"></param>
        /// <param name="depth">recusrion depth</param>
        /// <param name="threadId">actually just the index into the total</param>
        /// <param name="bits">since nodes gets changed on recursion, track the original total number of bits here.</param>
        static void SplitWorkRecursive(UInt64[] total, int nodes, int cliques, UInt64 add, UInt64 startIndex, int depth, int threadId, int bits)
        {
            // The values count up following the binomial coefficient
            var secondStart = nCk(nodes - 1, cliques);

            if (depth >= Global.MaxDepth)
            {
                Global.AddWorkers(2);

                Task.Factory.StartNew(() => Worker(total, nodes - 1, cliques, add, threadId*2, bits));
                Task.Factory.StartNew(() => Worker(total, nodes - 1, cliques - 1, add + (UInt64)((UInt64)1 << (nodes - 1)), threadId*2 + 1, bits));
            }
            else
            {
                SplitWorkRecursive(total, nodes - 1, cliques, add, startIndex, depth + 1, threadId * 2, bits);
                SplitWorkRecursive(total, nodes - 1, cliques - 1, add + (UInt64)((UInt64)1 << (nodes - 1)), startIndex + secondStart, depth + 1, threadId * 2 + 1, bits);
            }
        }

        static void Worker(UInt64[] total, int nodes, int cliques, UInt64 add, int threadId, int bits)
        {
            UInt64 familyCount = 0;

            UInt64 val = AllOnes(cliques) + add;
            var values = new List<UInt64>();

            values.Add(val);

            bool stillMoreWork = true;
            int maxCountTilMerge = 131072; // arbitrary

            var locky = new object();

            // since incrementing is relatively fast, do that a bunch of times.
            // After accumulating a lot of values, sort out which ones are valid.
            // Then repeat until done.

            Func<UInt64, UInt64> getIdAction = null;

            if (Global.Reflect)
            {
                getIdAction = x =>
                {
                    var id = GetId(x, bits);
                    var r = GetId(Reverse(x, bits), bits);

                    return Math.Min(id, r);
                };
            }
            else
            {
                getIdAction = x =>
                {
                    return GetId(x, bits);
                };
            }

            while (stillMoreWork)
            {
                for (int i=0; stillMoreWork && i < maxCountTilMerge; i++)
                {
                    if (Increment(val, nodes, out val))
                    {
                        values.Add(val);
                    }
                    else
                    {
                        stillMoreWork = false;
                    }
                }

                Parallel.ForEach(values,
                () => // Initializer
                {
                    return (UInt64)0;
                },
                (item, state, subtotal) => // Loop body
                {
                    var id = getIdAction(item);

                    if (id == item)
                    { 
                        return subtotal + 1;
                    }

                    return subtotal;
                },
                (subtotal) => // Accumulator
                {
                    lock (locky)
                    {
                        familyCount += subtotal;
                    }
                });

                values.Clear();
            }

            total[threadId] += familyCount;

            Global.RemoveWorkers(1);
        }

        /// <summary>
        /// Returns a value with the number of bits specified set.
        /// </summary>
        /// <param name="n">Number of bits to set.</param>
        /// <returns>Sets the specified least most significant bits.</returns>
        public static UInt64 AllOnes(int n)
        {
            UInt64 result = UInt64.MaxValue;

            return result >> (64 - n);
        }

        /// <summary>
        /// Increments to the next value with the same number of bits.
        /// </summary>
        /// <param name="n">Value to increment.</param>
        /// <param name="bits">Number of bits in the value that are set.</param>
        /// <param name="incremented">Next value. Can be same as input. If not incremented, will be the same as input.</param>
        /// <returns>True if it was possible to increment, false otherwise.</returns>
        public static bool Increment(UInt64 n, int bits, out UInt64 incremented)
        {
            int startPos = 0;
            int carry = 0;
            int index = 0;

            UInt64 copy = n;
            incremented = n;

            for (index = 0; index < bits; index++)
            {
                if ((copy & 0x01) == 0)
                {
                    startPos = index;
                    break;
                }

                copy = copy >> 1;
            }

            // Easy case. The digit can be shifted.
            if (startPos > 0)
            {
                incremented = incremented | ((UInt64)1 << startPos);
                incremented = (~((UInt64)((UInt64)1 << (startPos - 1)))) & incremented;
                return true;
            }

            // Else, the hard case. Flip and count the digits that need to be carried.
            for (index = startPos; index < bits; index++)
            {
                if ((copy & 0x01) == 0) 
                {
                    incremented = incremented | ((UInt64)1 << index);
                    carry++;
                }
                else
                {
                    break;
                }

                copy = copy >> 1;
            }

            // Travel down the array until it's possible to carry.
            for (index = carry + 1; index <= bits; index++)
            {
                if ((copy & 0x01) == 0)
                {
                    startPos = index - 1;
                    break;
                }

                copy = copy >> 1;
            }

            // If the end of the array was reached, it's not possible to
            // do anything. (max value for increment or min value for decrement)
            if (index > bits)
            {
                incremented = n;
                return false;
            }

            // Flip the digits "supplying" the carry
            incremented = incremented | ((UInt64)1 << startPos);
            incremented = (~((UInt64)((UInt64)1 << (startPos - 1)))) & incremented;

            // clear the digits that are carried down.
            for (index = startPos - 2; carry > 0; carry--, index--)
            {
                incremented = (~((UInt64)((UInt64)1 << index))) & incremented;
            }

            return true;
        }

        /// <summary>
        /// Rotates a value right once.
        /// </summary>
        /// <param name="n"></param>
        /// <param name="bits">Max number of times to rotate; also will rotate such that this is treated as the most significant bit.</param>
        /// <returns>The input value rotated right by one bit</returns>
        public static UInt64 RotateRight(UInt64 n, int bits)
        {
            UInt64 firstBit = n & 0x01;

            return (UInt64)(n >> 1) | ((UInt64)firstBit << (bits - 1));
        }

        public static UInt64 Reverse(UInt64 n, int bits)
        {
            UInt64 reverse = 0;

            for (int i=0; i<bits; i++)
            {
                UInt64 bit = (UInt64)(n & 0x01);
                reverse |= bit << (bits - i - 1);

                n = n >> 1;
            }

            return reverse;
        }

        /// <summary>
        /// Rotates a value through all possibilities and returns the smallest value seen.
        /// </summary>
        /// <param name="n">Value to find id for.</param>
        /// <param name="bits">Max number of times to rotate; also will rotate such that this is treated as the most significant bit.</param>
        /// <returns>The smallest value seen.</returns>
        public static UInt64 GetId(UInt64 n, int bits)
        {
            UInt64 id = n;

            for (int i=0; i<=bits; i++)
            {
                n = RotateRight(n, bits);
                id = Math.Min(id, n);
            }

            return id;
        }

        /// <summary>
        /// Binomial coefficient. n choose k.
        /// </summary>
        /// <param name="n"></param>
        /// <param name="k"></param>
        /// <returns></returns>
        public static UInt64 nCk(int n, int k)
        {
            UInt64 result = 1;

            System.Numerics.BigInteger b = 1;

            for (int i=n, count=0; count<k; count++,i--)
            {
                b *= i;
            }

            for (int i=k; i>1; i--)
            {
                b /= i;
            }

            result = UInt64.Parse(b.ToString());

            return result;
        }
    }
}
